import json
import logging
import sys
from json.decoder import JSONDecodeError
from typing import Any

from httpx import Response

LONG_PAYLOAD_VALUE_LIMIT = 300


def check_response(
    res: Response,
    expected_status_code: int | list[int] = 200,
    redact_long_payload: bool = False,
) -> list[Any] | dict[str, Any] | str | int | float | bool:
    """
    Check the validity of the http response from fractal server

    If the status code of the response is not one of the expected values, print
    the error to stderr and terminate with exit status 1.
    Some errors (422 errors due for failed validation of request body) are
    handled in a specific way, to make their error message more readable.

    Args:
        res: Response from `fractal-server`.
        expected_status_code: Expected status code(s).
        redact_long_payload: If `True`, redact payload values of more than
            `LONG_PAYLOAD_VALUE_LIMIT` characters.

    Returns:
        The output of `res.json()`.
    """

    try:
        data = res.json()
    except JSONDecodeError:
        data = {}

    # Also allow a list of expected status codes
    if isinstance(expected_status_code, list):
        expected_status_codes = expected_status_code
    else:
        expected_status_codes = [expected_status_code]

    logging.debug(f"\nResponse status code:\n    {res.status_code}")
    if res.status_code not in expected_status_codes:
        logging.error(f"Server returned {res.status_code}")

        # The following block relies on private methods, and it may fail for
        # unexpected reasons (e.g. it is now aware of the difference between
        # 'application/json' and 'multipart/form-data' requests, and it will
        # fail for non-jons requests). For this reason it is within a
        # broad-scope try/except block.
        try:
            logging.error(
                f"Original request: {res._request.method} {res._request.url}"
            )
            payload = res._request._content.decode("utf-8")
            if redact_long_payload and len(payload) > 0:
                payload_dict = json.loads(payload)
                for key, value in payload_dict.items():
                    if len(str(value)) > LONG_PAYLOAD_VALUE_LIMIT:
                        payload_dict[key] = "[value too long - redacted]"
                payload = json.dumps(payload_dict)
            logging.error(f"Original payload: {payload}")
        except Exception:
            logging.info("Could not display original-request information.")

        error_msg = data

        # Detect whether the error is due to failed request-body validation,
        # and make the error message more readable
        if (
            res.status_code == 422
            and isinstance(data, dict)
            and list(data.keys()) == ["detail"]
            and isinstance(data["detail"], list)
            and len(data["detail"]) == 1
            and isinstance(data["detail"][0], dict)
            and set(data["detail"][0].keys()) == {"msg", "type", "loc"}
        ):
            msg = data["detail"][0]["msg"]
            _type = data["detail"][0]["type"]
            loc = data["detail"][0]["loc"]
            error_msg = f"\n\tmsg: {msg}\n\ttype: {_type}\n\tloc: {loc}"

        logging.error(f"Server error message: {error_msg}\n")
        logging.error("Terminating.\n")
        sys.exit(1)

    return data



import json
import sys
from json.decoder import JSONDecodeError
from pathlib import Path

from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response


def _read_ssh_settings_json(ssh_settings_json: str) -> dict:
    """
    Read, validate and return as a dict the user's ssh-settings json file
    """
    ssh_settings_json_path = Path(ssh_settings_json)
    if not ssh_settings_json_path.exists():
        sys.exit(f"Invalid {ssh_settings_json=}. File does not exist.")
    with ssh_settings_json_path.open("r") as f:
        try:
            ssh_settings = json.load(f)
        except JSONDecodeError:
            sys.exit(f"{ssh_settings_json_path} is not a valid JSON.")
    __ALLOWED_KEYS__ = (
        "ssh_host",
        "ssh_username",
        "ssh_private_key_path",
        "ssh_tasks_dir",
        "ssh_jobs_dir",
    )
    settings = dict()
    for key, value in ssh_settings.items():
        if key in __ALLOWED_KEYS__:
            settings[key] = value
        else:
            sys.exit(f"Invalid {key=} in {ssh_settings_json=}.")

    return settings


def user_register(
    client: AuthClient,
    *,
    new_email: str,
    new_password: str,
    slurm_user: str | None = None,
    project_dir: str | None = None,
    username: str | None = None,
    ssh_settings_json: str | None = None,
    superuser: bool = False,
    verified: bool = True,  # TODO: this is not currently exposed in the CLI
    batch: bool = False,
) -> Interface:
    new_user = dict(
        email=new_email,
        password=new_password,
    )

    if username:
        new_user["username"] = username

    new_settings = dict()
    if slurm_user:
        new_settings["slurm_user"] = slurm_user
    if project_dir:
        new_settings["project_dir"] = project_dir
    if ssh_settings_json is not None:
        ssh_settings = _read_ssh_settings_json(ssh_settings_json)
        new_settings.update(ssh_settings)

    res = client.post("auth/register/", json=new_user)
    user_data = check_response(res, expected_status_code=201)

    if superuser or verified:
        patch_payload = dict(is_superuser=superuser, is_verified=verified)
        user_id = user_data["id"]
        res = client.patch(
            f"auth/users/{user_id}/",
            json=patch_payload,
        )
        user_data = check_response(res, expected_status_code=200)

    user_id = user_data["id"]
    if new_settings == {}:
        res = client.get(f"auth/users/{user_id}/settings/")
        user_settings = check_response(res, expected_status_code=200)
    else:
        res = client.patch(
            f"auth/users/{user_id}/settings/",
            json=new_settings,
        )
        user_settings = check_response(res, expected_status_code=200)

    if batch:
        return Interface(retcode=0, data=user_data["id"])
    else:
        user_data_with_settings = dict(settings=user_settings, **user_data)
        return Interface(retcode=0, data=user_data_with_settings)


def user_list(client: AuthClient) -> Interface:
    res = client.get("auth/users/")
    users = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=users)


def user_show(client: AuthClient, *, user_id: str) -> Interface:
    res = client.get(f"auth/users/{user_id}/")
    user = check_response(res, expected_status_code=200)
    user_id = user["id"]
    res = client.get(f"auth/users/{user_id}/settings/")
    user_settings = check_response(res, expected_status_code=200)
    user_with_settings = dict(settings=user_settings, **user)
    return Interface(retcode=0, data=user_with_settings)


def user_edit(
    client: AuthClient,
    *,
    user_id: str,
    new_email: str | None = None,
    new_password: str | None = None,
    new_username: str | None = None,
    new_slurm_user: str | None = None,
    new_project_dir: str | None = None,
    new_ssh_settings_json: str | None = None,
    make_superuser: bool = False,
    remove_superuser: bool = False,
    make_verified: bool = False,
    remove_verified: bool = False,
) -> Interface:
    user_update = dict()
    settings_update = dict()
    if new_email is not None:
        if (make_verified is False) and (remove_verified is False):
            # Since `fastapi-users` sets `is_verified` to `False` each time the
            # email is updated, we force the user to make explicit whether the
            # account is verified or not.
            return Interface(
                retcode=1,
                data=(
                    "Cannot use `--new-email` without `--make-verified` or "
                    "`--remove-verified`"
                ),
            )
        user_update["email"] = new_email
    if new_password is not None:
        user_update["password"] = new_password
    if make_superuser:
        user_update["is_superuser"] = True
    if remove_superuser:
        user_update["is_superuser"] = False
    if make_verified:
        user_update["is_verified"] = True
    if remove_verified:
        user_update["is_verified"] = False
    if new_username is not None:
        user_update["username"] = new_username
    if new_slurm_user is not None:
        settings_update["slurm_user"] = new_slurm_user
    if new_project_dir is not None:
        settings_update["project_dir"] = new_project_dir
    if new_ssh_settings_json is not None:
        ssh_settings = _read_ssh_settings_json(new_ssh_settings_json)
        settings_update.update(ssh_settings)

    res = client.patch(f"auth/users/{user_id}/", json=user_update)
    new_user = check_response(res, expected_status_code=200)

    if new_email is not None:
        # Since `fastapi-users` sets `is_verified` to `False` each time the
        # email is updated, we set `is_verified` as specified by the user.
        res = client.patch(
            f"auth/users/{user_id}/",
            json=dict(is_verified=user_update["is_verified"]),
        )
        new_user = check_response(res, expected_status_code=200)

    if settings_update == {}:
        res = client.get(f"auth/users/{user_id}/settings/")
        user_settings = check_response(res, expected_status_code=200)
    else:
        res = client.patch(
            f"auth/users/{user_id}/settings/",
            json=settings_update,
        )
        user_settings = check_response(res, expected_status_code=200)

    new_user_with_settings = dict(settings=user_settings, **new_user)
    return Interface(retcode=0, data=new_user_with_settings)


def user_set_groups(
    client: AuthClient, *, user_id: int, group_ids: list[int]
) -> Interface:
    res = client.post(
        f"auth/users/{user_id}/set-groups/",
        json=dict(group_ids=group_ids),
    )
    user = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=user)


def user_whoami(
    client: AuthClient, *, batch: bool, viewer_paths: bool = False
) -> Interface:
    res = client.get("auth/current-user/")
    user = check_response(res, expected_status_code=200)

    if batch:
        return Interface(retcode=0, data=user["id"])

    res = client.get("auth/current-user/settings/")
    user_settings = check_response(res, expected_status_code=200)
    user_with_settings = dict(**user, settings=user_settings)

    if viewer_paths:
        res = client.get("auth/current-user/allowed-viewer-paths/")
        returned_viewer_paths = check_response(res, expected_status_code=200)
        return Interface(
            retcode=0,
            data=dict(
                **user_with_settings,
                viewer_paths=returned_viewer_paths,
            ),
        )
    else:
        return Interface(retcode=0, data=user_with_settings)



import json
import logging
import os
from pathlib import Path
from zipfile import ZipFile

from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response


def get_job(
    client: AuthClient,
    *,
    project_id: int,
    job_id: int,
    batch: bool = False,
) -> Interface:
    """
    Query the status of a workflow-execution job
    """

    res = client.get(f"api/v2/project/{project_id}/job/{job_id}/")
    job = check_response(res, expected_status_code=200)
    if batch:
        return Interface(retcode=0, data=job["status"])
    else:
        return Interface(retcode=0, data=job)


def get_job_list(
    client: AuthClient, *, project_id: int, batch: bool = False
) -> Interface:
    res = client.get(f"api/v2/project/{project_id}/job/")
    jobs = check_response(res, expected_status_code=200)

    if batch:
        job_ids = " ".join(str(job["id"]) for job in jobs)
        return Interface(retcode=0, data=job_ids)
    else:
        return Interface(retcode=0, data=jobs)


def get_job_logs(
    client: AuthClient,
    *,
    project_id: int,
    job_id: int,
    output_folder: str,
) -> Interface:
    # Check that output_folder does not already exist
    if Path(output_folder).exists():
        return Interface(
            retcode=1, data=f"ERROR: {output_folder=} already exists"
        )

    # Send request to server
    res = client.get(f"api/v2/project/{project_id}/job/{job_id}/download/")

    # NOTE: We cannot use our default check_response here, because res._content
    # is binary. Therefore we check the status code by hand
    if res.status_code != 200:
        logging.error(f"Server returned {res.status_code}")
        logging.error(
            f"Original request: {res._request.method} {res._request.url}"
        )
        logging.error(
            f"Original payload: {res._request._content.decode('utf-8')}"
        )
        logging.error("Terminating.\n")
        exit(1)

    # Check the content-type entry in the response headers
    content_type = res.headers["content-type"]
    expected_content_type = "application/x-zip-compressed"
    if content_type != expected_content_type:
        logging.error(
            f"Unexpected {content_type=} in headers of server "
            f"response, instead of {expected_content_type=}"
        )
        logging.error(
            f"Original request: {res._request.method} {res._request.url}"
        )
        logging.error(
            f"Original payload: {res._request._content.decode('utf-8')}"
        )
        logging.error("Terminating.\n")
        exit(1)

    # Write response into a temporary zipped file
    zipped_archive_path = output_folder + "_tmp.zip"
    with open(zipped_archive_path, "wb") as f:
        f.write(res.content)

    # Unzip the log archive
    unzipped_archived_path = output_folder
    os.mkdir(unzipped_archived_path)
    with ZipFile(zipped_archive_path, mode="r") as zipfile:
        zipfile.extractall(path=unzipped_archived_path)

    # Remove zipped temporary file
    os.unlink(zipped_archive_path)

    return Interface(retcode=0, data=f"Logs downloaded to {output_folder=}")


def stop_job(client: AuthClient, *, project_id: int, job_id: int) -> Interface:
    """
    Stop a workflow-execution job
    """

    res = client.get(f"api/v2/project/{project_id}/job/{job_id}/stop/")
    check_response(res, expected_status_code=202)
    return Interface(
        retcode=0, data="Correctly called the job-stopping endpoint"
    )


def job_submit(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    dataset_id: int,
    first_task_index: int | None = None,
    last_task_index: int | None = None,
    worker_init: str | None = None,
    attribute_filters_json: str | None = None,
    type_filters_json: str | None = None,
    batch: bool = False,
) -> Interface:
    job_submit = dict()
    # Prepare JobV2 object, without None attributes
    if worker_init is not None:
        job_submit["worker_init"] = worker_init
    if first_task_index is not None:
        job_submit["first_task_index"] = first_task_index
    if last_task_index is not None:
        job_submit["last_task_index"] = last_task_index

    if attribute_filters_json is not None:
        with Path(attribute_filters_json).open("r") as f:
            job_submit["attribute_filters"] = json.load(f)
    if type_filters_json is not None:
        with Path(type_filters_json).open("r") as f:
            job_submit["type_filters"] = json.load(f)

    # Prepare query parameters
    query_parameters = f"workflow_id={workflow_id}" f"&dataset_id={dataset_id}"

    res = client.post(
        (f"api/v2/project/{project_id}/job/" f"submit/?{query_parameters}"),
        json=job_submit,
    )
    job_read = check_response(res, expected_status_code=202)

    if batch:
        return Interface(retcode=0, data=job_read["id"])
    else:
        return Interface(retcode=0, data=job_read)



from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response


def post_dataset(
    client: AuthClient,
    *,
    project_id: int,
    dataset_name: str,
    zarr_dir: str | None = None,
    batch: bool = False,
) -> Interface:
    """
    Arguments:
        project_id: ID of project to add the new dataset to
        dataset_name: Name of new dataset
        zarr_dir:
    """
    dataset = dict(name=dataset_name)
    if zarr_dir is not None:
        dataset["zarr_dir"] = zarr_dir

    res = client.post(
        f"api/v2/project/{project_id}/dataset/",
        json=dataset,
    )
    new_dataset = check_response(res, expected_status_code=201)
    if batch:
        return Interface(retcode=0, data=new_dataset["id"])
    else:
        return Interface(retcode=0, data=new_dataset)


def patch_dataset(
    client: AuthClient,
    *,
    project_id: int,
    dataset_id: int,
    new_name: str | None = None,
) -> Interface:
    # Prepare payload
    dataset_update = {}
    if new_name is not None:
        dataset_update["name"] = new_name

    res = client.patch(
        (f"api/v2/project/{project_id}/" f"dataset/{dataset_id}/"),
        json=dataset_update,
    )
    data = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=data)


def get_dataset(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> Interface:
    res = client.get(f"api/v2/project/{project_id}/dataset/{dataset_id}/")
    return Interface(retcode=0, data=res.json())


def delete_dataset(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> Interface:
    res = client.delete(f"api/v2/project/{project_id}/dataset/{dataset_id}/")
    check_response(res, expected_status_code=204)
    return Interface(retcode=0, data="")



import logging

from httpx import Client


logging.getLogger("httpx").setLevel(logging.WARNING)


def debug_request(verb: str, url: str, **kwargs):
    body = kwargs.get("json")
    log = f"\nFractal Client sending HTTP request to:\n    {verb} {url}"
    if body is not None:
        log += "\nRequest body:\n" + "\n".join(
            [f"    {k}: {v}" for k, v in body.items()]
        )
    logging.debug(log)


class AuthenticationError(ValueError):
    pass


class AuthClient:
    def __init__(
        self,
        *,
        fractal_server: str,
        username: str | None,
        password: str | None,
        token: str | None,
    ):
        self.fractal_server = fractal_server.rstrip("/")
        self.auth = None
        self.client = None
        self.username = username
        self.password = password
        self.token = token

    def __enter__(self):
        self.client = Client()
        if self.token is None:
            self.token = self._get_fresh_token()

        return self

    def __exit__(self, *args):
        self.client.close()

    def _get_fresh_token(self) -> str:
        res = self.client.post(
            f"{self.fractal_server}/auth/token/login/",
            data=dict(
                username=self.username,
                password=self.password,
            ),
        )
        if res.status_code != 200:
            data = res.text
            raise AuthenticationError(
                f"Error at {res.request.url}.\n"
                f"Status code: {res.status_code}.\n"
                f"Response data: {data}.\n"
            )
        raw_token = res.json()
        return raw_token["access_token"]

    @property
    def auth_headers(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self.token}"}

    def _get_url(self, relative_url: str) -> str:
        relative_url_no_leading_slash = relative_url.lstrip("/")
        return f"{self.fractal_server}/{relative_url_no_leading_slash}"

    def get(self, relative_url: str):
        url = self._get_url(relative_url)
        debug_request("GET", url)
        return self.client.get(url=url, headers=self.auth_headers)

    def post(self, relative_url: str, **kwargs):
        url = self._get_url(relative_url)
        debug_request("POST", relative_url, **kwargs)
        return self.client.post(url=url, headers=self.auth_headers, **kwargs)

    def patch(self, relative_url: str, **kwargs):
        url = self._get_url(relative_url)
        debug_request("PATCH", relative_url, **kwargs)
        return self.client.patch(url=url, headers=self.auth_headers, **kwargs)

    def delete(self, relative_url: str):
        url = self._get_url(relative_url)
        debug_request("DELETE", url)
        return self.client.delete(url=url, headers=self.auth_headers)



import json
from typing import Any


class Interface:
    def __init__(self, retcode: int, data: Any) -> None:
        self.retcode = retcode
        self.data = data

    def show(self, *args, **kwargs):
        if isinstance(self.data, (dict, list)):
            print(json.dumps(self.data, indent=2, sort_keys=True))
        else:
            print(self.data)



import json
import logging
import sys

from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response
from ._aux_task_caching import FractalCacheError
from ._aux_task_caching import get_task_id_from_cache
from ._aux_task_caching import refresh_task_cache


def get_task_list(client: AuthClient) -> Interface:
    task_list = refresh_task_cache(client=client)
    return Interface(retcode=0, data=task_list)


def post_task(
    client: AuthClient,
    *,
    name: str,
    task_type: str | None = None,
    batch: bool = False,
    command_non_parallel: str | None = None,
    command_parallel: str | None = None,
    version: str | None = None,
    meta_non_parallel: str | None = None,
    meta_parallel: str | None = None,
    args_schema_non_parallel: str | None = None,
    args_schema_parallel: str | None = None,
    args_schema_version: str | None = None,
    private: bool = False,
) -> Interface:
    task = dict(name=name)
    if task_type:
        task["type"] = task_type
    if command_non_parallel:
        task["command_non_parallel"] = command_non_parallel
    if command_parallel:
        task["command_parallel"] = command_parallel
    if version:
        task["version"] = version
    if meta_non_parallel:
        with open(meta_non_parallel) as f:
            task["meta_non_parallel"] = json.load(f)
    if meta_parallel:
        with open(meta_parallel) as f:
            task["meta_parallel"] = json.load(f)
    if args_schema_parallel:
        with open(args_schema_parallel) as f:
            task["args_schema_parallel"] = json.load(f)
    if args_schema_non_parallel:
        with open(args_schema_non_parallel) as f:
            task["args_schema_non_parallel"] = json.load(f)
    if args_schema_version:
        task["args_schema_version"] = args_schema_version
    is_private = "?private=true" if private else ""

    res = client.post(f"api/v2/task/{is_private}", json=task)
    new_task = check_response(res, expected_status_code=201)

    if batch:
        return Interface(retcode=0, data=str(new_task["id"]))
    else:
        return Interface(retcode=0, data=new_task)


def patch_task(
    client: AuthClient,
    *,
    id: int | None = None,
    name: str | None = None,
    version: str | None = None,
    command_non_parallel: str | None = None,
    command_parallel: str | None = None,
    input_types: str | None = None,
    output_types: str | None = None,
) -> Interface:
    if id:
        if version:
            logging.error(
                "Too many arguments: cannot provide both `id` and `version`."
            )
            sys.exit(1)
    else:
        try:
            id = get_task_id_from_cache(
                client=client, task_name=name, version=version
            )
        except FractalCacheError as e:
            print(e)
            sys.exit(1)

    task_update = {}
    if command_non_parallel:
        task_update["command_non_parallel"] = command_non_parallel
    if command_parallel:
        task_update["command_parallel"] = command_parallel
    if input_types:
        with open(input_types) as f:
            task_update["input_types"] = json.load(f)
    if output_types:
        with open(output_types) as f:
            task_update["output_types"] = json.load(f)

    res = client.patch(f"api/v2/task/{id}/", json=task_update)
    new_task = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=new_task)



from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response


def post_project(
    client: AuthClient,
    *,
    name: str,
    batch: bool = False,
) -> Interface:
    # Prepare a ProjectCreate request body
    payload = dict(name=name)
    # Send API request
    res = client.post("api/v2/project/", json=payload)
    project = check_response(res, expected_status_code=201)
    if batch:
        return Interface(retcode=0, data=project["id"])
    else:
        return Interface(retcode=0, data=project)


def get_project_list(client: AuthClient) -> Interface:
    res = client.get("api/v2/project/")
    projects = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=projects)


def get_project(client: AuthClient, *, project_id: int) -> Interface:
    res = client.get(f"api/v2/project/{project_id}/")
    project = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=project)


def delete_project(client: AuthClient, *, project_id: int) -> Interface:
    res = client.delete(f"api/v2/project/{project_id}/")
    check_response(res, expected_status_code=204)
    return Interface(retcode=0, data="")


def patch_project(
    client: AuthClient,
    *,
    project_id: int,
    new_name: str | None = None,
) -> Interface:
    project_update = {}
    if new_name:
        project_update["name"] = new_name

    res = client.patch(f"api/v2/project/{project_id}/", json=project_update)
    new_project = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=new_project)



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>
Marco Franzon <marco.franzon@exact-lab.it>
Tommaso Comparin <tommaso.comparin@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import logging
import sys

from httpx import ConnectError

from . import cmd
from .authclient import AuthClient
from .authclient import AuthenticationError
from .config import settings
from .interface import Interface
from .parser import parser_main


def _verify_authentication_branch(
    *,
    username: str | None,
    password: str | None,
    token_path: str | None,
) -> None:
    """
    Fail if credentials are not either username&password or token

    Arguments:
        username: Username
        password: Password
        token_path: Path of token
    """
    which_parameters_are_set = (
        bool(username),
        bool(password),
        bool(token_path),
    )
    valid_cases = (
        (True, True, False),
        (False, False, True),
    )
    if which_parameters_are_set not in valid_cases:
        msg = (
            "Invalid authentication credentials. "
            "You should either set username&password or the token path.\n\n"
            "You can set these variables in multiple ways "
            "(see `fractal --help`):\n"
            "  1. Through command-line arguments.\n"
            "  2. Through environment variables.\n"
            "  3. Through environment variables in a `.fractal.env` file.\n"
        )
        raise ValueError(msg)


def handle(cli_args: list[str] = sys.argv) -> Interface:
    args = parser_main.parse_args(cli_args[1:])

    # Set logging level
    if args.debug is True:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    show_args = "\n".join(
        [
            f"    {k}: {v}"
            if not (k == "password" and v is not None)
            else "    password: ***"
            for k, v in args.__dict__.items()
        ]
    )
    logging.debug(f"\nArguments:\n{show_args}")

    if args.cmd is not None:
        handler = getattr(cmd, args.cmd)
    else:
        parser_main.print_help()
        sys.exit(1)

    try:
        # Make a copy of vars(args), and remove cmd (which is not a relevant
        # argument for functions called with **kwargs)
        kwargs = vars(args).copy()
        kwargs.pop("cmd")
        fractal_server = (
            kwargs.pop("fractal_server") or settings.FRACTAL_SERVER
        )
        logging.debug(f"Fractal server URL: {fractal_server}")
        if fractal_server is None:
            return Interface(
                data=(
                    "Missing argument: You should set the "
                    "fractal-server URL (see `fractal --help`)."
                ),
                retcode=1,
            )

        if args.cmd == "version":
            interface = handler(fractal_server, **kwargs)
        else:
            # Extract (and remove) credentials-related variables from kwargs
            username = kwargs.pop("user") or settings.FRACTAL_USER
            password = kwargs.pop("password") or settings.FRACTAL_PASSWORD
            token_path = (
                kwargs.pop("token_path") or settings.FRACTAL_TOKEN_PATH
            )
            try:
                _verify_authentication_branch(
                    username=username,
                    password=password,
                    token_path=token_path,
                )
            except ValueError as e:
                return Interface(data=str(e), retcode=1)
            # Read token from file
            if token_path is not None:
                try:
                    with open(token_path) as f:
                        token = f.read().strip("\n")
                except Exception as e:
                    msg = (
                        f"Could not read token from {token_path=}.\n"
                        f"Original error:\n{str(e)}"
                    )
                    return Interface(data=msg, retcode=1)

            else:
                token = None

            with AuthClient(
                fractal_server=fractal_server,
                username=username,
                password=password,
                token=token,
            ) as client:
                interface = handler(client, **kwargs)
    except AuthenticationError as e:
        return Interface(retcode=1, data=e.args[0])
    except ConnectError as e:
        return Interface(
            retcode=1,
            data=(
                f"ConnectError at {e.request.url}\n"
                f"Original error: '{e.args[0]}'\n"
                f"Hint: is {settings.FRACTAL_SERVER} alive?"
            ),
        )
    except Exception as e:
        return Interface(retcode=1, data=str(e))

    return interface


def main():
    interface = handle()
    interface.show()
    sys.exit(interface.retcode)



from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response


def group_list(
    client: AuthClient,
    *,
    user_ids: bool = False,
    batch: bool = False,
):
    query_params = "?user_ids=true" if user_ids else ""
    res = client.get(f"auth/group/{query_params}")
    data = check_response(res, expected_status_code=200)
    if batch:
        return Interface(
            retcode=0, data=" ".join([str(d["id"]) for d in data])
        )
    else:
        return Interface(retcode=0, data=data)


def group_get(client: AuthClient, *, group_id: int):
    res = client.get(f"auth/group/{group_id}/")
    data = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=data)


def group_new(
    client: AuthClient,
    *,
    name: str,
    viewer_paths: list[str] | None = None,
    batch: bool = False,
):
    request_body = dict(name=name)
    if viewer_paths is not None:
        request_body["viewer_paths"] = viewer_paths

    res = client.post(
        "auth/group/",
        json=request_body,
    )
    data = check_response(res, expected_status_code=201)
    if batch:
        return Interface(retcode=0, data=data["id"])
    else:
        return Interface(retcode=0, data=data)


def group_update(
    client: AuthClient,
    *,
    group_id: int,
    new_viewer_paths: list[str],
):
    res = client.patch(
        f"auth/group/{group_id}/",
        json=dict(viewer_paths=new_viewer_paths),
    )
    data = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=data)


def group_add_user(
    client: AuthClient,
    *,
    group_id: int,
    user_id: int,
):
    res = client.post(f"auth/group/{group_id}/add-user/{user_id}/")
    data = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=data)


def group_remove_user(
    client: AuthClient,
    *,
    group_id: int,
    user_id: int,
):
    res = client.post(f"auth/group/{group_id}/remove-user/{user_id}/")
    data = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=data)



from httpx import Client
from httpx import ConnectError

from ..authclient import AuthClient
from ..interface import Interface
from ._dataset import delete_dataset
from ._dataset import get_dataset
from ._dataset import patch_dataset
from ._dataset import post_dataset
from ._group import group_add_user
from ._group import group_get
from ._group import group_list
from ._group import group_new
from ._group import group_remove_user
from ._group import group_update
from ._job import get_job
from ._job import get_job_list
from ._job import get_job_logs
from ._job import job_submit
from ._job import stop_job
from ._project import delete_project
from ._project import get_project
from ._project import get_project_list
from ._project import patch_project
from ._project import post_project
from ._task import get_task_list
from ._task import patch_task
from ._task import post_task
from ._task_collection import show_task_group_activity
from ._task_collection import task_collect_custom
from ._task_collection import task_collect_pip
from ._user import user_edit
from ._user import user_list
from ._user import user_register
from ._user import user_set_groups
from ._user import user_show
from ._user import user_whoami
from ._workflow import delete_workflow
from ._workflow import delete_workflowtask
from ._workflow import get_workflow
from ._workflow import get_workflow_list
from ._workflow import patch_workflow
from ._workflow import patch_workflowtask
from ._workflow import post_workflow
from ._workflow import post_workflowtask
from ._workflow import workflow_export
from ._workflow import workflow_import
from fractal_client import __VERSION__


class NoCommandError(ValueError):
    pass


def get_kwargs(_parameters, _kwargs):
    return {k: _kwargs.get(k) for k in _parameters if k in _kwargs}


def project(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> Interface:
    if subcmd == "new":
        parameters = ["name"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = post_project(client, batch=batch, **function_kwargs)
    elif subcmd == "show":
        parameters = ["project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = get_project(client, **function_kwargs)
    elif subcmd == "list":
        iface = get_project_list(client)
    elif subcmd == "edit":
        parameters = [
            "project_id",
            "new_name",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = patch_project(client, **function_kwargs)
    elif subcmd == "add-dataset":
        parameters = [
            "project_id",
            "dataset_name",
            "zarr_dir",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = post_dataset(client, batch=batch, **function_kwargs)
    elif subcmd == "delete":
        parameters = ["project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = delete_project(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command 'project {subcmd}' not found")

    return iface


def dataset(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> Interface:
    if subcmd == "show":
        parameters = ["project_id", "dataset_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = get_dataset(client, **function_kwargs)
    elif subcmd == "edit":
        parameters = [
            "project_id",
            "dataset_id",
            "new_name",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = patch_dataset(client, **function_kwargs)
    elif subcmd == "delete":
        parameters = ["project_id", "dataset_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = delete_dataset(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command 'dataset {subcmd}' not found")
    return iface


def task(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> Interface:
    if subcmd == "list":
        iface = get_task_list(client)
    elif subcmd == "collect":
        parameters = [
            "package",
            "package_version",
            "python_version",
            "package_extras",
            "pre_pinned_dependency",
            "post_pinned_dependency",
            "private",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = task_collect_pip(client, batch=batch, **function_kwargs)
    elif subcmd == "collect-custom":
        parameters = [
            "label",
            "python_interpreter",
            "manifest",
            "version",
            "package_name",
            "package_root",
            "private",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = task_collect_custom(client, batch=batch, **function_kwargs)
    elif subcmd == "check-collection":
        parameters = [
            "task_group_activity_id",
            "include_logs",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = show_task_group_activity(client, **function_kwargs)
    elif subcmd == "new":
        parameters = [
            "name",
            "task_type",
            "version",
            "command_non_parallel",
            "command_parallel",
            "meta_non_parallel",
            "meta_parallel",
            "args_schema_non_parallel",
            "args_schema_parallel",
            "args_schema_version",
            "private",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = post_task(client, batch=batch, **function_kwargs)
    elif subcmd == "edit":
        parameters = [
            "id",
            "name",
            "version",
            "command_non_parallel",
            "command_parallel",
            "input_types",
            "output_types",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = patch_task(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command 'task {subcmd}' not found")
    return iface


def workflow(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> Interface:
    if subcmd == "show":
        parameters = ["project_id", "workflow_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = get_workflow(client, **function_kwargs)
    elif subcmd == "new":
        parameters = ["name", "project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = post_workflow(client, batch=batch, **function_kwargs)
    elif subcmd == "list":
        parameters = ["project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = get_workflow_list(client, batch=batch, **function_kwargs)
    elif subcmd == "edit":
        parameters = ["project_id", "workflow_id", "new_name"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = patch_workflow(client, **function_kwargs)
    elif subcmd == "delete":
        parameters = ["project_id", "workflow_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = delete_workflow(client, **function_kwargs)
    elif subcmd == "add-task":
        parameters = [
            "project_id",
            "workflow_id",
            "task_id",
            "task_name",
            "task_version",
            "order",
            "args_non_parallel",
            "args_parallel",
            "type_filters",
            "meta_parallel",
            "meta_non_parallel",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = post_workflowtask(client, batch=batch, **function_kwargs)
    elif subcmd == "edit-task":
        parameters = [
            "project_id",
            "workflow_id",
            "workflow_task_id",
            "type_filters",
            "args_non_parallel",
            "args_parallel",
            "meta_parallel",
            "meta_non_parallel",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = patch_workflowtask(client, **function_kwargs)
    elif subcmd == "rm-task":
        parameters = ["project_id", "workflow_id", "workflow_task_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = delete_workflowtask(client, **function_kwargs)
    elif subcmd == "import":
        parameters = ["project_id", "json_file", "workflow_name"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = workflow_import(client, batch=batch, **function_kwargs)
    elif subcmd == "export":
        parameters = ["project_id", "workflow_id", "json_file"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = workflow_export(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command 'workflow {subcmd}' not found")
    return iface


def job(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> Interface:
    if subcmd == "list":
        parameters = ["project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = get_job_list(client, batch=batch, **function_kwargs)
    elif subcmd == "show":
        parameters = ["project_id", "job_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = get_job(client, batch=batch, **function_kwargs)
    elif subcmd == "download-logs":
        parameters = ["project_id", "job_id", "output_folder"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = get_job_logs(client, **function_kwargs)
    elif subcmd == "stop":
        parameters = ["project_id", "job_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = stop_job(client, **function_kwargs)
    elif subcmd == "submit":
        parameters = [
            "project_id",
            "workflow_id",
            "dataset_id",
            "worker_init",
            "first_task_index",
            "last_task_index",
            "attribute_filters_json",
            "type_filters_json",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = job_submit(client, batch=batch, **function_kwargs)
    else:
        raise NoCommandError(f"Command 'job {subcmd}' not found")
    return iface


def version(fractal_server: str, **kwargs) -> Interface:
    with Client() as client:
        try:
            res = client.get(f"{fractal_server}/api/alive/")
            server_version = res.json()["version"]
            server_str = f"  url: {fractal_server}  version: {server_version}"
        except ConnectError:
            server_str = f"  Connection to '{fractal_server}' refused"

        return Interface(
            retcode=0,
            data=(
                f"Fractal client\n  version: {__VERSION__}\n"
                f"Fractal server:\n{server_str}"
            ),
        )


def user(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> Interface:
    if subcmd == "register":
        parameters = [
            "new_email",
            "new_password",
            "slurm_user",
            "project_dir",
            "username",
            "ssh_settings_json",
            "superuser",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = user_register(client, batch=batch, **function_kwargs)
    elif subcmd == "list":
        iface = user_list(client)
    elif subcmd == "show":
        parameters = ["user_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = user_show(client, **function_kwargs)
    elif subcmd == "edit":
        parameters = [
            "user_id",
            "new_email",
            "new_password",
            "new_username",
            "new_slurm_user",
            "new_project_dir",
            "new_ssh_settings_json",
            "make_superuser",
            "remove_superuser",
            "make_verified",
            "remove_verified",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = user_edit(client, **function_kwargs)
    elif subcmd == "set-groups":
        parameters = [
            "user_id",
            "group_ids",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = user_set_groups(client, **function_kwargs)
    elif subcmd == "whoami":
        parameters = ["viewer_paths"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = user_whoami(client, **function_kwargs, batch=batch)
    else:
        raise NoCommandError(f"Command 'user {subcmd}' not found")

    return iface


def group(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> Interface:
    if subcmd == "list":
        parameters = ["user_ids"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = group_list(client, batch=batch, **function_kwargs)
    elif subcmd == "get":
        parameters = ["group_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = group_get(client, **function_kwargs)
    elif subcmd == "new":
        parameters = ["name", "viewer_paths"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = group_new(client, batch=batch, **function_kwargs)
    elif subcmd == "update":
        parameters = ["group_id", "new_viewer_paths"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = group_update(client, **function_kwargs)
    elif subcmd == "add-user":
        parameters = ["group_id", "user_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = group_add_user(client, **function_kwargs)
    elif subcmd == "remove-user":
        parameters = ["group_id", "user_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = group_remove_user(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command 'group {subcmd}' not found")

    return iface



import json
import logging
import sys
from pathlib import Path

from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response
from ._aux_task_caching import FractalCacheError
from ._aux_task_caching import get_task_id_from_cache


def post_workflow(
    client: AuthClient, *, name: str, project_id: int, batch: bool = False
) -> Interface:
    workflow = dict(
        name=name,
    )
    res = client.post(
        f"api/v2/project/{project_id}/workflow/",
        json=workflow,
    )
    workflow = check_response(res, expected_status_code=201)
    if batch:
        return Interface(retcode=0, data=workflow["id"])
    else:
        return Interface(retcode=0, data=workflow)


def get_workflow_list(
    client: AuthClient, *, project_id: int, batch: bool = False
) -> Interface:
    res = client.get(f"api/v2/project/{project_id}/workflow/")
    workflow_list = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=workflow_list)


def delete_workflow(
    client: AuthClient, *, project_id: int, workflow_id: int
) -> Interface:
    res = client.delete(f"api/v2/project/{project_id}/workflow/{workflow_id}/")
    check_response(res, expected_status_code=204)
    return Interface(retcode=0, data="")


def get_workflow(
    client: AuthClient, *, project_id: int, workflow_id: int
) -> Interface:
    res = client.get(f"api/v2/project/{project_id}/workflow/{workflow_id}/")
    workflow = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=workflow)


def post_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    type_filters: str | None = None,
    args_non_parallel: str | None = None,
    args_parallel: str | None = None,
    meta_non_parallel: str | None = None,
    meta_parallel: str | None = None,
    task_id: int | None = None,
    task_name: str | None = None,
    task_version: str | None = None,
    batch: bool = False,
    order: int | None = None,
) -> Interface:
    if task_id:
        if task_version:
            logging.error(
                "Too many arguments: cannot provide both "
                "`task_id` and `task_version`."
            )
            sys.exit(1)
    else:
        try:
            task_id = get_task_id_from_cache(
                client=client, task_name=task_name, version=task_version
            )
        except FractalCacheError as e:
            print(e)
            sys.exit(1)

    if order is None:
        workflow_task = dict()
    else:
        workflow_task = dict(order=order)

    if type_filters is not None:
        with Path(type_filters).open("r") as f:
            type_filters = json.load(f)
            workflow_task["type_filters"] = type_filters

    if args_non_parallel is not None:
        with Path(args_non_parallel).open("r") as f:
            a_n_p = json.load(f)
            workflow_task["args_non_parallel"] = a_n_p

    if args_parallel is not None:
        with Path(args_parallel).open("r") as f:
            a_p = json.load(f)
            workflow_task["args_parallel"] = a_p

    if meta_non_parallel is not None:
        with Path(meta_non_parallel).open("r") as f:
            m_n_p = json.load(f)
            workflow_task["meta_non_parallel"] = m_n_p

    if meta_parallel is not None:
        with Path(meta_parallel).open("r") as f:
            m_p = json.load(f)
            workflow_task["meta_parallel"] = m_p

    res = client.post(
        (
            f"api/v2/project/{project_id}/"
            f"workflow/{workflow_id}/wftask/"
            f"?{task_id=}"
        ),
        json=workflow_task,
    )
    workflow_task = check_response(res, expected_status_code=201)

    if batch:
        return Interface(retcode=0, data=str(workflow_task["id"]))
    else:
        return Interface(retcode=0, data=workflow_task)


def patch_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    workflow_task_id: int,
    type_filters: str | None = None,
    args_non_parallel: str | None = None,
    args_parallel: str | None = None,
    meta_non_parallel: str | None = None,
    meta_parallel: str | None = None,
) -> Interface:
    payload = {}
    if type_filters is not None:
        with Path(type_filters).open("r") as f:
            filters = json.load(f)
            payload["type_filters"] = filters

    if args_non_parallel is not None:
        with Path(args_non_parallel).open("r") as f:
            a_n_p = json.load(f)
            payload["args_non_parallel"] = a_n_p

    if args_parallel is not None:
        with Path(args_parallel).open("r") as f:
            a_p = json.load(f)
            payload["args_parallel"] = a_p

    if meta_non_parallel is not None:
        with Path(meta_non_parallel).open("r") as f:
            m_n_p = json.load(f)
            payload["meta_non_parallel"] = m_n_p

    if meta_parallel is not None:
        with Path(meta_parallel).open("r") as f:
            m_p = json.load(f)
            payload["meta_parallel"] = m_p

    res = client.patch(
        (
            f"api/v2/project/{project_id}/"
            f"workflow/{workflow_id}/wftask/{workflow_task_id}/"
        ),
        json=payload,
    )
    workflow_task = check_response(res, expected_status_code=200)

    return Interface(retcode=0, data=workflow_task)


def delete_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    workflow_task_id: int,
) -> Interface:
    res = client.delete(
        f"api/v2/project/{project_id}/"
        f"workflow/{workflow_id}/wftask/{workflow_task_id}/"
    )
    check_response(res, expected_status_code=204)
    return Interface(retcode=0, data="")


def patch_workflow(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    new_name: str,
) -> Interface:
    workflow_update = dict(name=new_name)

    res = client.patch(
        f"api/v2/project/{project_id}/workflow/{workflow_id}/",
        json=workflow_update,
    )
    new_workflow = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=new_workflow)


def workflow_import(
    client: AuthClient,
    *,
    project_id: int,
    json_file: str,
    workflow_name: str | None = None,
    batch: bool = False,
) -> Interface:
    with Path(json_file).open("r") as f:
        workflow = json.load(f)

    if workflow_name is not None:
        workflow["name"] = workflow_name

    res = client.post(
        f"api/v2/project/{project_id}/workflow/import/",
        json=workflow,
    )
    wf_read = check_response(res, expected_status_code=201)

    if batch:
        datastr = f"{wf_read['id']}"
        for wftask in wf_read["task_list"]:
            datastr += f" {wftask['id']}"
        return Interface(retcode=0, data=datastr)
    else:
        return Interface(retcode=0, data=wf_read)


def workflow_export(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    json_file: str,
) -> Interface:
    res = client.get(
        (f"api/v2/project/{project_id}/" f"workflow/{workflow_id}/export/"),
    )
    workflow = check_response(res, expected_status_code=200)

    with Path(json_file).open("w") as f:
        json.dump(workflow, f, indent=2)
    return Interface(
        retcode=0, data=f"Workflow {workflow_id} exported at {json_file}"
    )



import json
import logging
import sys
from pathlib import Path
from typing import Literal

from fractal_client.authclient import AuthClient
from fractal_client.interface import Interface
from fractal_client.response import check_response


def _process_pinned_dependencies(
    pinned_dependencies: list[str],
    label: Literal["pre", "post"],
) -> dict[str, str]:
    """
    Transform `["a=1", "b=2"]` into `{"a": "1", "b": "2"}`.
    """
    for pin in pinned_dependencies:
        if len(pin.split("=")) != 2:
            logging.error(
                f"Invalid pin: {pin}.\nPins must be written as "
                f"'--{label}-pinned-dependency PACKAGE_NAME=PACKAGE_VERSION'"
            )
            sys.exit(1)

    return json.dumps(
        {
            _name: _version
            for _name, _version in (p.split("=") for p in pinned_dependencies)
        }
    )


def task_collect_pip(
    client: AuthClient,
    *,
    package: str,
    package_version: str | None = None,
    python_version: str | None = None,
    package_extras: str | None = None,
    pre_pinned_dependency: list[str] | None = None,
    post_pinned_dependency: list[str] | None = None,
    private: bool = False,
    batch: bool = False,
) -> Interface:
    # Construct TaskCollectPip object
    task_collect = dict()
    if package_version:
        task_collect["package_version"] = package_version
    if python_version:
        task_collect["python_version"] = python_version
    if package_extras:
        task_collect["package_extras"] = package_extras
    if pre_pinned_dependency:
        task_collect[
            "pre_pinned_package_versions"
        ] = _process_pinned_dependencies(
            pre_pinned_dependency,
            "pre",
        )
    if post_pinned_dependency:
        task_collect[
            "post_pinned_package_versions"
        ] = _process_pinned_dependencies(
            post_pinned_dependency,
            "post",
        )

    is_private = "?private=true" if private else ""
    endpoint_url = f"api/v2/task/collect/pip/{is_private}"
    if package.endswith(".whl"):
        with open(package, "rb") as f:
            file = {
                "file": (
                    Path(package).name,
                    f.read(),
                    "application/zip",
                )
            }
        res = client.post(
            endpoint_url,
            data=task_collect,
            files=file,
        )
    else:
        task_collect["package"] = package
        res = client.post(
            endpoint_url,
            data=task_collect,
        )
    task_group_activity = check_response(res, expected_status_code=202)
    if batch:
        return Interface(retcode=0, data=task_group_activity["id"])
    else:
        return Interface(retcode=0, data=task_group_activity)


def task_collect_custom(
    client: AuthClient,
    *,
    label: str,
    python_interpreter: str,
    manifest: str,
    version: str | None = None,
    package_name: str | None = None,
    package_root: str | None = None,
    private: bool = False,
    batch: bool = False,
) -> Interface:
    try:
        with open(manifest) as f:
            manifest_dict = json.load(f)
    except FileNotFoundError as e:
        raise FileNotFoundError(
            f"Fractal Client cannot find the file {manifest}. "
            "Note that the file must be on the same machine where Fractal "
            f"Client is running.\nOriginal error: {e}."
        )

    task_collect = dict(
        label=label,
        python_interpreter=python_interpreter,
        manifest=manifest_dict,
    )
    if version:
        task_collect["version"] = version
    if package_name:
        task_collect["package_name"] = package_name
    if package_root:
        task_collect["package_root"] = package_root
    is_private = "?private=true" if private else ""

    res = client.post(
        f"api/v2/task/collect/custom/{is_private}",
        json=task_collect,
    )

    task_list = check_response(
        res, expected_status_code=201, redact_long_payload=True
    )

    if batch:
        task_ids = [str(task["id"]) for task in task_list]
        return Interface(retcode=0, data=" ".join(task_ids))
    else:
        return Interface(retcode=0, data=task_list)


def show_task_group_activity(
    client: AuthClient,
    *,
    task_group_activity_id: int,
    include_logs: bool,
) -> Interface:
    res = client.get(f"api/v2/task-group/activity/{task_group_activity_id}/")
    task_group_activity = check_response(res, expected_status_code=200)

    # Remove key-value pairs with None value
    if include_logs is False:
        task_group_activity["log"] = None

    return Interface(retcode=0, data=task_group_activity)
